#ifndef BCArray_H
#define BCArray_H

#include "BoundaryConditions.h"
#include "basics/container/CbArray3D.h"

#include <typeinfo>

#include <boost/serialization/serialization.hpp>
#include <boost/shared_ptr.hpp>

typedef boost::shared_ptr<BoundaryConditions> BCClassPtr;

class BCArray3D;
typedef boost::shared_ptr<BCArray3D> BCArray3DPtr;

class BCArray3D
{
public:
   //////////////////////////////////////////////////////////////////////////
   BCArray3D();
   //////////////////////////////////////////////////////////////////////////
   BCArray3D(std::size_t nx1, std::size_t nx2, std::size_t nx3);
   //////////////////////////////////////////////////////////////////////////
   BCArray3D(std::size_t nx1, std::size_t nx2, std::size_t nx3, int val);
   //////////////////////////////////////////////////////////////////////////
   virtual ~BCArray3D();
   //////////////////////////////////////////////////////////////////////////
   inline std::size_t getNX1() const;
   //////////////////////////////////////////////////////////////////////////
   inline std::size_t getNX2() const;
   //////////////////////////////////////////////////////////////////////////
   inline std::size_t getNX3() const;
   //////////////////////////////////////////////////////////////////////////
   void resize(std::size_t nx1, std::size_t nx2, std::size_t nx3);
   //////////////////////////////////////////////////////////////////////////
   void resize(std::size_t nx1, std::size_t nx2, std::size_t nx3, int val);
   //////////////////////////////////////////////////////////////////////////
   bool validIndices(std::size_t x1, std::size_t x2, std::size_t x3)  const;
   //////////////////////////////////////////////////////////////////////////
   inline bool hasBC(std::size_t x1, std::size_t x2, std::size_t x3)  const;
   //////////////////////////////////////////////////////////////////////////
   void setBC(std::size_t x1, std::size_t x2, std::size_t x3, BCClassPtr const& bc);
   //////////////////////////////////////////////////////////////////////////
   inline int getBCVectorIndex(std::size_t x1, std::size_t x2, std::size_t x3) const;
   //////////////////////////////////////////////////////////////////////////
   inline const BCClassPtr getBC(std::size_t x1, std::size_t x2, std::size_t x3) const;
   //////////////////////////////////////////////////////////////////////////
   inline BCClassPtr getBC(std::size_t x1, std::size_t x2, std::size_t x3);
   //////////////////////////////////////////////////////////////////////////
   void setSolid(std::size_t x1, std::size_t x2, std::size_t x3);
   //////////////////////////////////////////////////////////////////////////
   inline bool isSolid(std::size_t x1, std::size_t x2, std::size_t x3) const;
   //////////////////////////////////////////////////////////////////////////
   void setFluid(std::size_t x1, std::size_t x2, std::size_t x3);
   //////////////////////////////////////////////////////////////////////////
   //true : FLUID or BC
   //false: UNDEFINED or SOLID
   inline bool isFluid(std::size_t x1, std::size_t x2, std::size_t x3) const;
   //////////////////////////////////////////////////////////////////////////
   inline bool isFluidWithoutBC(std::size_t x1, std::size_t x2, std::size_t x3) const;
   //////////////////////////////////////////////////////////////////////////
   inline bool isUndefined(std::size_t x1, std::size_t x2, std::size_t x3) const;
   //////////////////////////////////////////////////////////////////////////
   void setUndefined(std::size_t x1, std::size_t x2, std::size_t x3);
   //////////////////////////////////////////////////////////////////////////
   inline bool isInterfaceCF(std::size_t x1, std::size_t x2, std::size_t x3) const;
   //////////////////////////////////////////////////////////////////////////
   void setInterfaceCF(std::size_t x1, std::size_t x2, std::size_t x3);
   //////////////////////////////////////////////////////////////////////////
   inline bool isInterfaceFC(std::size_t x1, std::size_t x2, std::size_t x3) const;
   //////////////////////////////////////////////////////////////////////////
   void setInterfaceFC(std::size_t x1, std::size_t x2, std::size_t x3);
   //////////////////////////////////////////////////////////////////////////
   std::size_t getNumberOfSolidEntries() const;
   //////////////////////////////////////////////////////////////////////////
   std::size_t getNumberOfFluidEntries() const;
   //////////////////////////////////////////////////////////////////////////
   std::size_t getNumberOfFluidWithoutBCEntries() const;
   //////////////////////////////////////////////////////////////////////////
   std::size_t getNumberOfBCEntries() const;
   //////////////////////////////////////////////////////////////////////////
   std::size_t getNumberOfUndefinedEntries() const;
   //////////////////////////////////////////////////////////////////////////
   std::size_t getBCVectorSize() const;
   //////////////////////////////////////////////////////////////////////////
   std::string toString() const;
   //////////////////////////////////////////////////////////////////////////

   static const int SOLID;     
   static const int FLUID;     
   static const int INTERFACECF; 
   static const int INTERFACEFC; 
   static const int UNDEFINED; 

private:
   //////////////////////////////////////////////////////////////////////////
   void deleteBCAndSetType(std::size_t x1, std::size_t x2, std::size_t x3, int type);
   //////////////////////////////////////////////////////////////////////////
   void deleteBC(std::size_t x1, std::size_t x2, std::size_t x3);

   friend class MPIIORestartCoProcessor;

   friend class boost::serialization::access;
   template<class Archive>
   void serialize(Archive & ar, const unsigned int version)
   {
      ar & bcindexmatrix;
      ar & bcvector;
      ar & indexContainer;
   }
protected:
   //////////////////////////////////////////////////////////////////////////
   //-1 solid // -2 fluid -...
   CbArray3D<int, IndexerX3X2X1> bcindexmatrix;
   std::vector<BCClassPtr> bcvector;
   std::vector<int> indexContainer;
};


//////////////////////////////////////////////////////////////////////////
inline std::size_t BCArray3D::getNX1() const { return bcindexmatrix.getNX1(); }
//////////////////////////////////////////////////////////////////////////
inline std::size_t BCArray3D::getNX2() const { return bcindexmatrix.getNX2(); }
//////////////////////////////////////////////////////////////////////////
inline std::size_t BCArray3D::getNX3() const { return bcindexmatrix.getNX3(); }
//////////////////////////////////////////////////////////////////////////
inline bool BCArray3D::hasBC(std::size_t x1, std::size_t x2, std::size_t x3)  const
{
   return bcindexmatrix(x1, x2, x3) >= 0;
}
//////////////////////////////////////////////////////////////////////////
inline int BCArray3D::getBCVectorIndex(std::size_t x1, std::size_t x2, std::size_t x3) const
{
   return bcindexmatrix(x1, x2, x3);
}
//////////////////////////////////////////////////////////////////////////
inline const BCClassPtr  BCArray3D::getBC(std::size_t x1, std::size_t x2, std::size_t x3) const
{
   int index = bcindexmatrix(x1, x2, x3);
   if (index < 0) return BCClassPtr(); //=> NULL Pointer

   return bcvector[index];
}
//////////////////////////////////////////////////////////////////////////
inline BCClassPtr BCArray3D::getBC(std::size_t x1, std::size_t x2, std::size_t x3)
{
   int index = bcindexmatrix(x1, x2, x3);
   if (index < 0) return BCClassPtr(); //=> NULL Pointer

   return bcvector[index];
}
//////////////////////////////////////////////////////////////////////////
inline bool BCArray3D::isSolid(std::size_t x1, std::size_t x2, std::size_t x3) const
{
   return bcindexmatrix(x1, x2, x3) == SOLID;
}
//////////////////////////////////////////////////////////////////////////
//true : FLUID or BC
//false: UNDEFINED or SOLID
inline bool BCArray3D::isFluid(std::size_t x1, std::size_t x2, std::size_t x3) const
{
   int tmp = bcindexmatrix(x1, x2, x3);
   return (tmp == FLUID || tmp >= 0);
}
//////////////////////////////////////////////////////////////////////////
inline bool BCArray3D::isFluidWithoutBC(std::size_t x1, std::size_t x2, std::size_t x3) const
{
   return bcindexmatrix(x1, x2, x3) == FLUID;
}
//////////////////////////////////////////////////////////////////////////
inline bool BCArray3D::isUndefined(std::size_t x1, std::size_t x2, std::size_t x3) const
{
   return bcindexmatrix(x1, x2, x3) == UNDEFINED;
}
//////////////////////////////////////////////////////////////////////////
inline bool BCArray3D::isInterfaceCF(std::size_t x1, std::size_t x2, std::size_t x3) const
{
   return bcindexmatrix(x1, x2, x3) == INTERFACECF;
}
//////////////////////////////////////////////////////////////////////////
inline bool BCArray3D::isInterfaceFC(std::size_t x1, std::size_t x2, std::size_t x3) const
{
   return bcindexmatrix(x1, x2, x3) == INTERFACEFC;
}

#endif 
